//====================================================================================//
//                                                                                    //
//                            Class Probability Vector                                //
//                                                                                    //
//====================================================================================//
//  This File:   vct2.C                     Language: C++  (xlC and CC)               //
//  Software:    PEPS 2002                   Advisor: Brigitte.Plateau@imag.fr        //
//  Error Codes: 2400                                                                 //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 13/feb/02                        by:     Anne.Benoit@imag.fr         //
//====================================================================================//
#include "peps.h"

      //--------------------------------------------------------------------------//
      //      Static variables to the Internal Functions of Multiply by ...       //
      //                                                                          //
      //   The following five variables are defined as static to reduce the cost  //
      // for allocation and release memory structures in the most sensible point  //
      // of the PEPS solvers.                                                     //
      //--------------------------------------------------------------------------//
VCT    VCT::t;       // temporary vector to store the intermediary results
rp   * VCT::z_in;    // input "slice" (a pointer to an array of vectors elements)
rp   * VCT::z_out;   // output "slice" (a pointer to an array of vectors elements)
aut_st VCT::g_st;    // a global state according to the current descriptor size

     //--------------------------------------------------------------------------//
     //       Multiply by a CND  -  Method A  -  'this' = 'v' * 'CND::cnd'       //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_A(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2403)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2404)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions

 for (a=0; a<CND::cnd.automata; a++)
   switch (CND::cnd.lf_type[a]) {
     case idnt_mat : add(v); // The RSS is the same
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.local_mat[a], v, a);
                     break;
     case func_mat : mfip(CND::cnd.local_mat[a], CND::cnd.state_map, v, a);
   }
 for (e=0; e<CND::cnd.events; e++) {
   a = CND::cnd.s_D_order[e].aut(0);
   switch (CND::cnd.sf_type[e][a]) {
     case idnt_mat : for (k=0; k<dimension; k++)
                       t.vec[k] = v.vec[k];
                     break;
     case null_mat : for (k=0; k<dimension; k++)
                       t.vec[k] = zero;
                     break;
     case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                     break;
     case func_mat : t.mfis(CND::cnd.synch_mat[e][a], CND::cnd.state_map, v, a);
   }

   for (i=1; i<CND::cnd.automata-1; i++) {
     a = CND::cnd.s_D_order[e].aut(i);
     switch (CND::cnd.sf_type[e][a]) {
       case idnt_mat : break;
       case null_mat : for (k=0; k<dimension; k++)
                         t.vec[k] = zero;
                       i=CND::cnd.automata-1;  // stops the loop
                       break;
       case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                       break;
       case func_mat : t.mfis(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a);
     }
   }
   a = CND::cnd.s_D_order[e].aut(CND::cnd.automata-1);
   switch (CND::cnd.sf_type[e][a]) {
     case idnt_mat : for (k=0; k<dimension; k++)
                       vec[k] += t.vec[k];
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.synch_mat[e][a], t, a);
                     break;
     case func_mat : mfip(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a);
   }
 }
}
   
     //--------------------------------------------------------------------------//
     //       Multiply by a CND  -  Method B  -  'this' = 'v' * 'CND::cnd'       //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_B(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2405)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2406)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions

 for (a=0; a<CND::cnd.automata; a++)
   switch (CND::cnd.lf_type[a]) {
     case idnt_mat : add(v); 
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.local_mat[a], v, a);
                     break;
     case func_mat : t.becomes(v, CND::cnd.l_B_order[a], CND::cnd.l_B_jumps[a]);
                     t.mfos(CND::cnd.local_mat[a], CND::cnd.state_map, t, a);
                     (*this) += t;
   }
 for (e=0; e<CND::cnd.events; e++) {
   t.becomes(v, CND::cnd.s_B_order[e], CND::cnd.s_B_jumps[e]);
   for (i=0; i<CND::cnd.automata; i++) {
     a = CND::cnd.s_B_order[e].aut(i);
     switch (CND::cnd.sf_type[e][a]) {
       case idnt_mat : break;
       case null_mat : for (k=0; k<dimension; k++)
                         t.vec[k] = zero;
                       i=CND::cnd.automata-1;  // stops the loop
                       break;
       case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                       break;
       case func_mat : t.mfos(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a);
     }
   }
 (*this) += t;
 }
}

     //--------------------------------------------------------------------------//
     //       Multiply by a CND  -  Method C  -  'this' = 'v' * 'CND::cnd'       //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
     //--------------------------------------------------------------------------//
     //       Multiply by a CND  -  Method C  -  'this' = 'v' * 'CND::cnd'       //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_C(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2407)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2408)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions

 for (a=0; a<CND::cnd.automata; a++)
   switch (CND::cnd.lf_type[a]) {
     case idnt_mat : add(v);
                     break;
     case null_mat : break;
     case cnst_mat : mcp(CND::cnd.local_mat[a], v, a);
                     break;
     case func_mat : t.becomes(v, CND::cnd.l_C_order[a], CND::cnd.l_C_jumps[a]);
                     t.mnps(CND::cnd.local_mat[a], CND::cnd.state_map, t, a,
                            CND::cnd.l_C_left[a], CND::cnd.l_C_right[a],
                            CND::cnd.l_C_fst[a]);
                     (*this) += t;
   }

 for (e=0; e<CND::cnd.events; e++) {
   //------------------------------
   t.get_CND();  // it copies the permutation info of the original vector ('v')
   //------------------------------
   a = CND::cnd.s_D_order[e].aut(0);
   switch (CND::cnd.sf_type[e][a]) {
     case idnt_mat : for (k=0; k<dimension; k++)
                       t.vec[k] = v.vec[k];
                     break;
     case null_mat : for (k=0; k<dimension; k++)
                       t.vec[k] = zero;
                     break;
     case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                     break;
     case func_mat : t.becomes(v, CND::cnd.s_C_order[e][a], CND::cnd.s_C_jumps[e][a]);
                     t.mnps(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a,
                     CND::cnd.s_C_left[e][a], CND::cnd.s_C_right[e][a],
                     CND::cnd.s_C_fst[e][a]);
   }
   for (i=1; i<CND::cnd.automata; i++) {
     a = CND::cnd.s_D_order[e].aut(i);
     switch (CND::cnd.sf_type[e][a]) {
       case idnt_mat : break;
       case null_mat : for (k=0; k<dimension; k++)
                         t.vec[k] = zero;
                       i=CND::cnd.automata-1;  // stops the loop
                       break;
       case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                       break;
       case func_mat : t.permut(CND::cnd.s_C_order[e][a], CND::cnd.s_C_jumps[e][a]);
                       t.mnps(CND::cnd.synch_mat[e][a], CND::cnd.state_map, t, a,
                       CND::cnd.s_C_left[e][a], CND::cnd.s_C_right[e][a],
                       CND::cnd.s_C_fst[e][a]);
     }
   }
 (*this) += t;
 }
}
 
 
     //--------------------------------------------------------------------------//
     //     Multiply by a CND  -  Mixed Method  -  'this' = 'v' * 'CND::cnd'     //
     //            this.positions == v.positions (call from mult_CND)            //
     //--------------------------------------------------------------------------//
void VCT::mult_CND_M(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2409)
  if (CND::cnd.ss() != dimension)
    Programming_Error("bad sized CND", 2410)
#endif
 int a;            // counter for the automata
 int e;            // counter for the synchronizing events
 int i;            // a counter for positions in a list of automata
 int k;            // a counter for vector positions

 for (a=0; a<CND::cnd.automata; a++) {
   switch (CND::cnd.l_type[a]) {
     case idnt_term : for (k=0; k<dimension; k++)
                        vec[k] += one;
                      break;
     case null_term : break;
     case cnst_term : mcp(CND::cnd.local_mat[a], v, a);
                      break;
     case A_term    : mfip(CND::cnd.local_mat[a], CND::cnd.state_map, v, a);
                      break;
     case B_term    : t.becomes(v, CND::cnd.l_B_order[a], CND::cnd.l_B_jumps[a]);
                      t.mfos(CND::cnd.local_mat[a], CND::cnd.state_map, t, a);
                      (*this) += t;
                      break;
     case C_term    : t.becomes(v, CND::cnd.l_C_order[a], CND::cnd.l_C_jumps[a]);
                      t.mnps(CND::cnd.local_mat[a], CND::cnd.state_map, t, a,
                             CND::cnd.l_C_left[a], CND::cnd.l_C_right[a],
                             CND::cnd.l_C_fst[a]);
                      (*this) += t;
   }
 }

 for (e=0; e<CND::cnd.events; e++) {
   //------------------------------
   t.get_CND();  // it copies the permutation info of the original vector ('v')
   //------------------------------
   switch (CND::cnd.s_type[e]) {
     case idnt_term : for (k=0; k<dimension; k++)
                        vec[k] += one;
                      break;
     case null_term : break;
     case cnst_term : 
     case A_term    : a = CND::cnd.s_D_order[e].aut(0);
                      switch (CND::cnd.sf_type[e][a]) {
                        case idnt_mat : for (k=0; k<dimension; k++)
                                          t.vec[k] = v.vec[k];
                                        break;
                        case null_mat : for (k=0; k<dimension; k++)
                                          t.vec[k] = zero;
                                        break;
                        case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                                        break;
                        case func_mat : t.mfis(CND::cnd.synch_mat[e][a],
                                               CND::cnd.state_map, v, a);
                      }
                      for (i=1; i<CND::cnd.automata-1; i++) {
                        a = CND::cnd.s_D_order[e].aut(i);
                        switch (CND::cnd.sf_type[e][a]) {
                          case idnt_mat : break;
                          case null_mat : for (k=0; k<dimension; k++)
                                            t.vec[k] = zero;
                                            i=CND::cnd.automata-1;  // stops the loop
                                          break;
                          case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                                          break;
                          case func_mat : t.mfis(CND::cnd.synch_mat[e][a],
                                                 CND::cnd.state_map, t, a);
                        }
                      }
                      a = CND::cnd.s_D_order[e].aut(CND::cnd.automata-1);
                      switch (CND::cnd.sf_type[e][a]) {
                        case idnt_mat : for (k=0; k<dimension; k++)
                                          vec[k] += t.vec[k];
                                        break;
                        case null_mat : break;
                        case cnst_mat : mcp(CND::cnd.synch_mat[e][a], t, a);
                                        break;
                        case func_mat : mfip(CND::cnd.synch_mat[e][a],
                                             CND::cnd.state_map, t, a);
                      }
                      break;
     case B_term    : t.becomes(v, CND::cnd.s_B_order[e], CND::cnd.s_B_jumps[e]);
                      for (i=0; i<CND::cnd.automata; i++) {
                        a = CND::cnd.s_B_order[e].aut(i);
                        switch (CND::cnd.sf_type[e][a]) {
                          case idnt_mat : break;
                          case null_mat : for (k=0; k<dimension; k++)
                                            t.vec[k] = zero;
                                          i=CND::cnd.automata-1;  // stops the loop
                                          break;
                          case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                                          break;
                          case func_mat : t.mfos(CND::cnd.synch_mat[e][a],
                                                 CND::cnd.state_map, t, a);
                        }
                      }
                      (*this) += t;
                      break;
     case C_term    : a = CND::cnd.s_D_order[e].aut(0);
                      switch (CND::cnd.sf_type[e][a]) {
                        case idnt_mat : for (k=0; k<dimension; k++)
                                          t.vec[k] = v.vec[k];
                                        break;
                        case null_mat : for (k=0; k<dimension; k++)
                                          t.vec[k] = zero;
                                        break;
                        case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], v, a);
                                        break;
                        case func_mat : t.becomes(v, CND::cnd.s_C_order[e][a],
                                                     CND::cnd.s_C_jumps[e][a]);
                                        t.mnps(CND::cnd.synch_mat[e][a],
                                               CND::cnd.state_map, t, a,
                                               CND::cnd.s_C_left[e][a],
                                               CND::cnd.s_C_right[e][a],
                                               CND::cnd.s_C_fst[e][a]);
                      }
                      for (i=1; i<CND::cnd.automata; i++) {
                        a = CND::cnd.s_D_order[e].aut(i);
                        switch (CND::cnd.sf_type[e][a]) {
                          case idnt_mat : break;
                          case null_mat : for (k=0; k<CND::cnd.state_space; k++)
                                            t.vec[k] = zero;
                                            i=CND::cnd.automata-1;  // stops the loop
                                          break;
                          case cnst_mat : t.mcs(CND::cnd.synch_mat[e][a], t, a);
                                          break;
                          case func_mat : t.permut(CND::cnd.s_C_order[e][a],
                                                   CND::cnd.s_C_jumps[e][a]);
                                          t.mnps(CND::cnd.synch_mat[e][a],
                                                 CND::cnd.state_map, t, a,
                                                 CND::cnd.s_C_left[e][a],
                                                 CND::cnd.s_C_right[e][a],
                                                 CND::cnd.s_C_fst[e][a]);
                        }
                      }
                      (*this) += t;
   }
 }
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //     It multiplies a vector 'v' by a normal factor without functions (non //
     // identity matrix in 'm') adding the result to the vector 'this'.          //
     //--------------------------------------------------------------------------//
void VCT::mcp(san_mat & m, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int idx, base = 0;                    // pointer to pick vector elements for the slices

 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   for (r=0; r<nright; r++) {                            // for the right sub-space:
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       z_in[i] = v.vec[idx];                             // gets the z_in slice
       idx += nright;                                    //
     }                                                   //
     m.left_mult(z_out, z_in);                           // z_out = z_in * matrix
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       vec[idx] += z_out[i];                             // stores the z_out slice
       idx += nright;                                    //
     }                                                   //
   }
   base += njump;                                      // pass to the next
 }                                                     // left-space block
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') adding the result to the vector 'this'.          //
     //                                                                          //
     // Obs: The function evaluations are performed inside the inner most loop   //
     // (no permutations are need).                                              //
     //--------------------------------------------------------------------------//
void VCT::mfip(san_mat & m, const st_map * sm, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int idx, base = 0;                    // pointer to pick vector elements for the slices

 g_st.reset();                                        // global state reset
 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   for (r=0; r<nright; r++) {                            // for the right sub-space:
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       z_in[i] = v.vec[idx];                             // gets the z_in slice
       idx += nright;                                    //
     }                                                   //
     m.jit_eval(g_st, a, sm);                                // evaluates the matrix
     m.left_mult(z_out, z_in);                           // z_out = z_in * matrix
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       vec[idx] += z_out[i];                             // stores the z_out slice
       idx += nright;                                    //
     }                                                   //
     g_st.next(v.sizes, a, Right);                       // changes the state right
   }
   g_st.next(v.sizes, a, Left);                        // changes the state right
   base += njump;                                      // pass to the next
 }                                                     // left-space block
}
     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //     It multiplies a vector 'v' by a normal factor without functions (non //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //--------------------------------------------------------------------------//
void VCT::mcs(san_mat & m, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int idx, base = 0;                    // pointer to pick vector elements for the slices

 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   for (r=0; r<nright; r++) {                            // for the right sub-space:
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       z_in[i] = v.vec[idx];                             // gets the z_in slice
       idx += nright;                                    //
     }                                                   //
     m.left_mult(z_out, z_in);                           // z_out = z_in * matrix
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       vec[idx] = z_out[i];                              // stores the z_out slice
       idx += nright;                                    //
     }                                                   //
   }
   base += njump;                                     // pass to the next
 }                                                    // left-space block
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //                                                                          //
     // Obs: The function evaluations are performed inside the inner most loop   //
     // (no permutations are need).                                              //
     //--------------------------------------------------------------------------//
void VCT::mfis(san_mat & m, const st_map * sm, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int pos       = v.p_order.pos(a);     // position of automaton 'a' in the list
 int idx, base = 0;                    // pointer to pick vector elements for the slices

 g_st.reset();                                        // global state reset
 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   for (r=0; r<nright; r++) {                            // for the right sub-space:
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       z_in[i] = v.vec[idx];                             // gets the z_in slice
       idx += nright;                                    //
     }                                                   //
     m.jit_eval(g_st, a, sm);                                // evaluates the matrix
     m.left_mult(z_out, z_in);                           // z_out = z_in * matrix
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       vec[idx] = z_out[i];                              // stores the z_out slice
       idx += nright;                                    //
     }                                                   //
     g_st.next(v.sizes, pos, Right);                     // pass to the next
   }                                                     // right-space block
   base += njump;                                     // pass to the next
   g_st.next(v.sizes, pos, Left);                     // left-space block
 }
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //                                                                          //
     // Obs: The function evaluations are performed outside the inner most loop  //
     // (permutations are need).                                                 //
     //--------------------------------------------------------------------------//
void VCT::mfos(san_mat & m, const st_map * sm, const VCT & v, const aut_id a)
{
 int i, ni     = v.sizes.sz(a);        // size of the automaton 'a'
 int l, nleft  = v.p_jumps.nleft(a);   // space size before the automaton 'a'
 int r, nright = v.p_jumps.nright(a);  // space size after the automaton 'a'
 int njump     = v.p_jumps.njump(a);   // 'nright' times 'ni'
 int pos       = v.p_order.pos(a);     // position of automaton 'a' in the list
 int idx, base = 0;                    // pointer to pick vector elements for the slices

 g_st.reset();                                        // global state reset
 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   m.jit_eval(g_st, a, sm);                               // evaluates the matrix
   for (r=0; r<nright; r++) {                            // for the right sub-space:
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       z_in[i] = v.vec[idx];                             // gets the z_in slice
       idx += nright;                                    //
     }                                                   //
     m.left_mult(z_out, z_in);                           // z_out = z_in * matrix
     idx = base + r;                                     //
     for (i=0; i<ni; i++) {                              //
       vec[idx] = z_out[i];                              // stores the z_out slice
       idx += nright;                                    //
     }                                                   //
   }
   base += njump;                                     // pass to the next
   g_st.next(v.p_order, v.sizes, pos, Left);          // left-space block
 }
}

     //--------------------------------------------------------------------------//
     //                  Internal Function of Multiply by a CND                  //
     //                                                                          //
     //      It multiplies a vector 'v' by a normal factor with functions (non   //
     // identity matrix in 'm') storing the result in the vector 'this'.         //
     //                                                                          //
     // Obs: The permutation of the input vector is performed before this        //
     // function is called, the function evaluation is performed outside the     //
     // inner most loop (no permutations are need in this function).             //
     //--------------------------------------------------------------------------//
void VCT::mnps(san_mat & m, const st_map * sm, const VCT & v, const aut_id a,
               const int nleft, const int nright, const aut_id pos)
{
 int i, ni = v.sizes.sz(a);    // size of the automaton 'a'
 int l; // nleft               // space size of the parameters of automaton 'a'
 int r; // nright              // space size of the non parameters of automaton 'a'
 int njump = nright * ni;      // 'nright' times 'ni'
        // pos                 // position of the first non parameter of 'a' in the list
 int base = 0;                 // pointer to pick vector elements for the slices

 g_st.reset();                                        // global state reset
 for (l=0; l<nleft; l++) {                            // for the left sub-space:
   m.jit_eval(g_st, a, sm);                               // evaluates the matrix
   for (r=0; r<nright; r++) {                            // for the right sub-space:
     m.left_mult(z_out, &(v.vec[base]));                 // z_out = slice * matrix
     for (i=0; i<ni; i++)                                //
       vec[base+i] = z_out[i];                           // stores the z_out slice
     base += ni;                                         //
   }
   g_st.next(v.p_order, v.sizes, pos, Left);          // pass to the next
 }                                                    // left-space block
}


      //--------------------------------------------------------------------------//
      //      Multiply by a CND  -  generic switch among methods A, B, C and M    //
      //--------------------------------------------------------------------------//
void VCT::mult_CND(const VCT & v)
{
  for (int k=0; k<dimension; k++)         // dot product of the diagonal
   vec[k] = v.vec[k] * CND::cnd.diagonal[k];

 switch(PRF::prf.method()) {                        // switch methods A, B, C or M
   case meth_A : mult_CND_A(v); break;
   case meth_B : mult_CND_B(v); break;
   case meth_C : mult_CND_C(v); break;
   case meth_M : mult_CND_M(v);
 }
}

     //--------------------------------------------------------------------------//
     //                Multiply by the inverse of the CND diagonal               //
     //--------------------------------------------------------------------------//
void VCT::solv_D()
{
 for (int k=0; k<dimension; k++)       // dot product of the inverse of
   vec[k] /= CND::cnd.diagonal[k];                //    the diagonal
}

     //--------------------------------------------------------------------------//
     // Preconditionned Multiplications (multiplications + solutions)            //
     //--------------------------------------------------------------------------//
     // Diagonal       - uses the diagonal of the 'CND::cnd' descriptor as       //
     //                  preconditionner.                                        //
     //                                          'this = (v * CND) * diag(CND)'  //
     //--------------------------------------------------------------------------//
void VCT::pre_diag(const VCT & v)
{
 mult_CND(v);                          // it performs 'this = v * CND'
 solv_D();                             // it performs 'this = (v * CND) * diag(CND)'
}
     //--------------------------------------------------------------------------//
     // Preconditionned Multiplications (multiplications + solutions)            //
     //--------------------------------------------------------------------------//
     // Polynomial     - uses the 'CND::cnd' as the preconditionner and          //
     //                  aproximates its inverse by the polynome:                //
     //                  'I + CND + CND^2 + CND^3 + ...'                         //
     //                              'this = v * CND * (I + CND + CND^2 + ...')  //
     //--------------------------------------------------------------------------//
void VCT::pre_plyn(const VCT & v)
{
 int n_terms = PRF::prf.poly_k();
 VCT zz(dimension);

 mult_CND(v);                          // it performs 'zz = this = v * CND'
 zz.copy(*this);
 for (int i=1; i<n_terms; i++) {       // for all polynomial terms
   mult_CND(*this);                       // it performs 'this *= CND'
   add(zz);                               // it performs 'this = (v * CND) + this'
 }
}
     //--------------------------------------------------------------------------//
     // Preconditionned Multiplications (multiplications + solutions)            //
     //--------------------------------------------------------------------------//
     // I-A Polynomial - uses the Identity - 'CND::cnd' as the preconditionner   //
     //                  and aproximates its inverse by the polynome:            //
     //                  'I + (I - CND) + (I - CND)^2 + (I - CND)^3 + ...'       //
     //                  'this = v * CND * (I + (I - CND) + (I - CND)^2 + ...')  //
     //--------------------------------------------------------------------------//
void VCT::pre_ipln(const VCT & v)
{
 int n_terms = PRF::prf.poly_k();
 int k;
 rp  buf;
 VCT zz(dimension);
 VCT ww(dimension);

 mult_CND(v);                             // it performs 'this = v * CND'
 for (k=0; k<dimension; k++)              // it performs 'zz = v * CND * (I - CND)'
   zz.vec[k] = v.vec[k] - vec[k];
 for (int i=0; i<n_terms; i++) {          // for all polynomial terms 
   ww.mult_CND(zz);                          // it performs 'ww = zz * CND'
   for (k=0; k<dimension; k++) {
     buf = zz.vec[k] - ww.vec[k];
     vec[k] += buf;                          // it performs 'this += zz * (I - CND)'
     zz.vec[k] = buf;                        // it performs 'zz = zz * (I - CND)'
   }
 }
}
     //--------------------------------------------------------------------------//
     // Preconditionned Multiplications (multiplications + solutions)            //
     //--------------------------------------------------------------------------//
     // Chebyshev      - uses ??????????????????????????                         //
     //                                    'this = v * (I + CND + CND^2 + ...')  //
     //--------------------------------------------------------------------------//
void VCT::pre_cheb(const VCT & v)
{
 int n_terms = PRF::prf.poly_k();
 VCT zz(dimension);

 mult_CND(v);                          // it performs 'this = v * CND'
 for (int i=1; i<n_terms; i++) {       // for all polynomial terms
   zz.copy(*this);                        // it copies 'x^(k-1)'
   mult_CND(*this);                       // it performs 'this *= CND'
   scale_add(zz, -one);                   // it performs 'this = (v * CND) + this'
 }
}
